"""Provides an example how to use dyscom layer get commands"""

import sys
import asyncio

from science_mode_4 import SerialPortConnection
from science_mode_4 import DeviceI24
from examples.utils.example_utils import ExampleUtils


async def main() -> int:
    """Main function"""

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceI24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get general layer to access general commands
    general = device.get_layer_general()
    # get dyscom layer to access dyscom commands
    dyscom = device.get_layer_dyscom()

    #####
    # dyscom layer provides a get command for device id, it is redundant to general layer device id
    # both values should be identical
    device_id = await dyscom.get_device_id()
    print(f"Device id from general layer: {general.device_id}, from dyscom layer: {device_id}")

    #####
    # dyscom layer provides a get command for firmware version, it is redundant to general layer firmware verion
    # both values should be identical
    firmware_version = await dyscom.get_firmware_version()
    print(f"Firmware version from general layer: {general.firmware_version}, from dyscom layer: {firmware_version}")

    #####
    # get command for file system status
    file_system_status = await dyscom.get_file_system_status()
    print(file_system_status)

    # close serial port connection
    connection.close()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides packet classes for dyscom get"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.utils.byte_builder import ByteBuilder
from .dyscom_types import DyscomGetType


class PacketDyscomGet(Packet):
    """Packet for dyscom get, use descendants of this class"""


    def __init__(self):
        super().__init__()
        self._command = Commands.DL_GET
        self._type = DyscomGetType.BATTERY


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_byte(self._type)
        return bb.get_bytes()


class PacketDyscomGetAck(PacketAck):
    """Packet for dyscom get acknowledge, use descendants of this class"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_GET_ACK
        self._result_error = ResultAndError.NO_ERROR
        self._type = DyscomGetType.BATTERY

        if not data is None:
            self._result_error = ResultAndError(data[0])
            self._type = DyscomGetType(data[1])


    def get_kind(self, data: bytes) -> int:
        """Get kind from data"""        
        # use type as kind
        return data[1]


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error


    @property
    def type(self) -> DyscomGetType:
        """Getter for type"""
        return self._type



"""Provides packet classes for dyscom get with type file system status"""

# from dataclasses import dataclass

from typing import NamedTuple
from science_mode_4.protocol.commands import Commands
from .dyscom_types import DyscomGetType
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class DyscomGetFileSystemStatusResult(NamedTuple):
    """Helper class for dyscom get with type file system status"""
    file_system_ready: bool
    used_size: int
    free_size: int


class PacketDyscomGetFileSystemStatus(PacketDyscomGet):
    """Packet for dyscom get with type file system status"""


    def __init__(self):
        super().__init__()
        self._command = Commands.DL_GET
        self._type = DyscomGetType.FILESYSTEM_STATUS
        self._kind = int(self._type)


class PacketDyscomGetAckFileSystemStatus(PacketDyscomGetAck):
    """Packet for dyscom get for type file system status acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.FILESYSTEM_STATUS)
        self._file_system_ready = False
        self._used_size = 0
        self._free_size = 0

        if not data is None:
            self._file_system_ready = bool(data[2])
            self._used_size = int.from_bytes(data[3:11], "little")
            self._free_size = int.from_bytes(data[11:19], "little")


    @property
    def file_system_ready(self) -> bool:
        """Getter for file system ready"""
        return self._file_system_ready


    @property
    def used_size(self) -> int:
        """Getter for used size"""
        return self._used_size


    @property
    def free_size(self) -> int:
        """Getter for free size"""
        return self._free_size



"""Provides class for ADS129x chip config register 2"""

from dataclasses import dataclass
from enum import IntEnum


class Ads129xWctChoppingScheme(IntEnum):
    """Represent ADS129xtype for WCT chopping scheme (configuration register 2, 0x02)"""
    CHOPPING_FREQUENCY_VARIES = 0
    CHOPPING_FREQUENCY_CONSTANT = 1


class Ads129xTestSignalSource(IntEnum):
    """Represent ADS129xtype for test signal source (configuration register 2, 0x02)"""
    EXTERNAL_TEST_SIGNAL = 0
    INTERNAL_TEST_SIGNAL = 1


class Ads129xTestSignalAmplitude(IntEnum):
    """Represent ADS129xtype for test signal amplitude (configuration register 2, 0x02)"""
    SINGLE_VREF = 0
    DOUBLE_VREF = 1


class Ads129xTestSignalFrequency(IntEnum):
    """Represent ADS129xtype for test signal frequency (configuration register 2, 0x02)"""
    PULSE_AT_2_21 = 0
    PULSE_AT_2_20 = 1
    PULSE_AT_DC = 3


@dataclass
class Ads129xConfigRegister2:
    """Describes config register 2 of ADS129x chip"""

    wct_chopping_scheme = Ads129xWctChoppingScheme.CHOPPING_FREQUENCY_VARIES
    test_signal_source = Ads129xTestSignalSource.EXTERNAL_TEST_SIGNAL
    test_signal_amplitude = Ads129xTestSignalAmplitude.SINGLE_VREF
    test_signal_frequency = Ads129xTestSignalFrequency.PULSE_AT_2_21


    def set_data(self, data: bytes):
        """Convert data to information"""
        tmp = data[0]
        self.wct_chopping_scheme = Ads129xWctChoppingScheme((tmp >> 5) & 0x01)
        self.test_signal_source = Ads129xTestSignalSource((tmp >> 4) & 0x01)
        self.test_signal_amplitude = Ads129xTestSignalAmplitude((tmp >> 2) & 0x01)
        self.test_signal_frequency = Ads129xTestSignalFrequency((tmp >> 0) & 0x03)


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        tmp = ((self.wct_chopping_scheme << 5) | (self.test_signal_source << 4) |
               (self.test_signal_amplitude << 2) | (self.test_signal_frequency << 0))
        return [tmp]



"""Provides a logger"""

import logging


class Logger:
    """Singleton for custom logger. By default a StreamHandler with a custom formatter is added
    and log level set to INFO"""

    _instance = None


    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Logger, cls).__new__(cls)
            # because __init__ would be called every time we use Logger, initialize everything here
            cls._logger = logging.getLogger("science_mode_4")
            cls._console_handler = None
            cls._initialize(cls)
        return cls._instance


    @property
    def logger(self) -> logging.Logger:
        """Getter for logger"""
        return self._logger


    def _initialize(self):
        self._logger.setLevel(logging.INFO)

        formatter = logging.Formatter("{asctime} - {levelname} - {message}", style="{", datefmt="%Y-%m-%d %H:%M")

        self._console_handler = logging.StreamHandler()
        self._console_handler.setFormatter(formatter)
        self._logger.addHandler(self._console_handler)



def logger() -> logging.Logger:
    """Shortcut to access custom logger"""
    return Logger().logger



"""Provides a ByteBuilder class for easier creation of a byte stream"""

from .bit_vector import BitVector


class ByteBuilder():
    """ByteBuilder class"""


    def __init__(self, data: int = 0, byte_count: int = 0):
        self._data = BitVector.init_from_int(data, byte_count * 8)


    def get_bit_from_position(self, bit_position: int, bit_count: int) -> int:
        """Returns bits starting with bit_position and a count of bit_count"""
        result = 0
        for x in range(bit_count):
            result |= (self._data[bit_position + x] << x)
        return result


    def append_value(self, value: int, byte_count: int, do_swap: bool):
        """Extends current data with byte_count bytes from values"""
        temp = range(byte_count)
        if do_swap:
            temp = reversed(temp)
        for x in temp:
            self._append_byte(value >> (x * 8))


    def append_byte(self, value: int):
        """Append a byte"""
        self._append_byte(value)


    def append_list(self, value: list[int]):
        """Extends current data with list of values (byte)"""
        for x in value:
            self._append_byte(x)


    def append_bytes(self, value: bytes):
        """Extends current data with value"""
        for x in value:
            self._append_byte(x)


    def extend_byte_builder(self, value: "ByteBuilder"):
        """Extends current data with value"""
        self._data.extend(value.get_bytes())


    def set_bit_to_position(self, value: int, bit_position: int, bit_count: int):
        """
        Set bits starting with bit_position and a count of bit_count to value
        This method extends data to make room for value
        """
        new_length = max(len(self._data), bit_position + bit_count)
        self._data.set_length(new_length)
        for x in range(bit_count):
            self._data[bit_position + x] = (value >> x) & 0x1


    def set_bytes_to_position(self, value: bytes, byte_position: int, byte_count: int):
        """
        Set bytes starting with byte_position and a count of byte_count to value
        This method extends data to make room for value
        """
        for x in range(byte_count):
            self.set_bit_to_position(value[x], (byte_position + x) * 8, 8)


    def swap(self, start: int, count: int):
        """Swap bytes by reversion order from start to start + count"""
        tmp = self.get_bytes()
        for x in range(count):
            self.set_bit_to_position(tmp[start + x], (start + count - x - 1) * 8, 8)


    def get_bytes(self) -> bytes:
        """Returns data as bytes"""
        return self._data.get_bytes()


    def clear(self):
        """Resets data"""
        self._data = BitVector()


    def __len__(self) -> int:
        return (len(self._data) + 7) // 8


    def __repr__(self) -> str:
        b = self.get_bytes()
        return f"length: {len(b)}, bytes: {b.hex(" ").upper()}"


    def __str__(self) -> str:
        b = self.get_bytes()
        return f"length: {len(b)}, bytes: {b.hex(" ").upper()}"


    def _append_byte(self, value: int):
        """Append value at the end of data, value is treated as byte"""
        start = len(self._data)
        self._data.set_length(start + 8)
        for x in range(8):
            self._data[start + x] = (value >> x) & 0x1



"""Provides classes for mid level Init"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.utils.byte_builder import ByteBuilder


class PacketMidLevelInit(Packet):
    """Packet for mid level Init"""


    def __init__(self):
        super().__init__()
        self._command = Commands.MID_LEVEL_INIT
        self._do_stop_on_all_errors = False


    @property
    def do_stop_on_all_errors(self) -> bool:
        """Getter for do stop on all errors"""
        return self._do_stop_on_all_errors


    @do_stop_on_all_errors.setter
    def do_stop_on_all_errors(self, do_stop_on_all_errors: bool):
        """Setter for do stop on all errors"""
        self._do_stop_on_all_errors = do_stop_on_all_errors


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_byte(1 if self._do_stop_on_all_errors else 0)
        return bb.get_bytes()


class PacketMidLevelInitAck(PacketAck):
    """Packet for mid level Init acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.MID_LEVEL_INIT_ACK
        self._result_error = ResultAndError.NO_ERROR

        if not data is None:
            self._result_error = ResultAndError(data[0])


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error



"""Provides an example how to use dyscom level layer to measure emg and bi and plot values"""

import asyncio

from science_mode_4 import DeviceI24
from science_mode_4 import Commands
from science_mode_4 import SerialPortConnection
from science_mode_4.dyscom.ads129x.ads129x_config_register_1 import Ads129xOutputDataRate, Ads129xPowerMode
from science_mode_4.dyscom.dyscom_get_operation_mode import PacketDyscomGetAckOperationMode
from science_mode_4.dyscom.dyscom_send_live_data import PacketDyscomSendLiveData
from science_mode_4.dyscom.dyscom_types import DyscomInitParams, DyscomPowerModulePowerType, DyscomPowerModuleType, DyscomSignalType
from science_mode_4.utils.logger import logger
from examples.utils.example_utils import ExampleUtils
from examples.utils.pyplot_utils import PyPlotHelper


async def main() -> int:
    """Main function"""

    plot_helper = PyPlotHelper({0: ["BI", "blue"]}, 250)

    # disable logger to increase performance
    logger().disabled = True

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceI24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get dyscom layer to call low level commands
    dyscom = device.get_layer_dyscom()

    # call enable measurement power module for measurement
    await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_ON)
    # call init with lowest sample rate (because of performance issues with plotting values)
    init_params = DyscomInitParams()
    init_params.signal_type = [DyscomSignalType.BI, DyscomSignalType.EMG_1]
    init_params.register_map_ads129x.config_register_1.output_data_rate = Ads129xOutputDataRate.HR_MODE_500_SPS__LP_MODE_250_SPS
    init_params.register_map_ads129x.config_register_1.power_mode = Ads129xPowerMode.LOW_POWER
    await dyscom.init(init_params)

    # start dyscom measurement
    await dyscom.start()

    for x in range(1000):
        # check operation mode from time to time
        if x % 100 == 0:
            dyscom.send_get_operation_mode()

        live_data_counter = 0
        while True:
            ack = dyscom.packet_buffer.get_packet_from_buffer(live_data_counter == 0)
            if ack:
                if ack.command == Commands.DL_GET_ACK:
                    om_ack: PacketDyscomGetAckOperationMode = ack
                    print(f"Operation mode {om_ack.operation_mode.name}")
                elif ack.command == Commands.DL_SEND_LIVE_DATA:
                    live_data_counter += 1

                    sld: PacketDyscomSendLiveData = ack
                    if sld.status_error:
                        print(f"SendLiveData status error {sld.samples}")
                        break

                    # reduce framerate further
                    if sld.number % 60 == 0:
                        plot_helper.append_value(0, sld.samples[0].value)
                        plot_helper.update()

            else:
                # print(f"Live data acknowledges per iteration {live_data_counter}")
                break

        await asyncio.sleep(0.01)

    # stop measurement
    await dyscom.stop()
    # turn power module off
    await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_OFF)

    # close serial port connection
    connection.close()

    return 0


if __name__ == "__main__":
    asyncio.run(main())



"""Provides packet classes for low level init"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.utils.byte_builder import ByteBuilder
from .low_level_types import LowLevelHighVoltageSource, LowLevelMode


class PacketLowLevelInit(Packet):
    """Packet for low level init"""


    def __init__(self):
        super().__init__()
        self._command = Commands.LOW_LEVEL_INIT
        self._mode = LowLevelMode.NO_MEASUREMENT
        self._high_voltage_source = LowLevelHighVoltageSource.STANDARD


    @property
    def mode(self) -> LowLevelMode:
        """Getter for mode"""
        return self._mode


    @mode.setter
    def mode(self, value: LowLevelMode):
        """Setter for mode"""
        self._mode = value


    @property
    def high_voltage_source(self) -> int:
        """Getter for high voltage source"""
        return self._high_voltage_source


    @high_voltage_source.setter
    def high_voltage_source(self, value: int):
        """Setter for high voltage source"""
        self._high_voltage_source = value


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.set_bit_to_position(0, 0, 1)
        bb.set_bit_to_position(self._high_voltage_source, 1, 3)
        bb.set_bit_to_position(self._mode, 4, 3)
        bb.set_bit_to_position(0, 7, 0)
        return bb.get_bytes()


class PacketLowLevelInitAck(PacketAck):
    """Packet for low level init acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.LOW_LEVEL_INIT_ACK
        self._result_error = ResultAndError.NO_ERROR

        if not data is None:
            self._result_error = ResultAndError(data[0])


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error



"""Provides all science mode command numbers"""


from enum import IntEnum


class Commands(IntEnum):
    """Enum with all commands"""

    UNDEFINED = -1

    GET_DEVICE_ID = 52
    GET_DEVICE_ID_ACK = 53
    RESET = 58
    RESET_ACK = 59
    GET_STIM_STATUS = 62
    GET_STIM_STATUS_ACK = 63
    GET_EXTENDED_VERSION = 68
    GET_EXTENDED_VERSION_ACK = 69
    GENERAL_ERROR = 66
    UNKNOWN_COMMAND = 67

    LOW_LEVEL_INIT = 0
    LOW_LEVEL_INIT_ACK = 1
    LOW_LEVEL_CHANNEL_CONFIG = 2
    LOW_LEVEL_CHANNEL_CONFIG_ACK = 3
    LOW_LEVEL_STOP = 4
    LOW_LEVEL_STOP_ACK = 5

    MID_LEVEL_INIT = 30
    MID_LEVEL_INIT_ACK = 31
    MID_LEVEL_UPDATE = 32
    MID_LEVEL_UPDATE_ACK = 33
    MID_LEVEL_STOP = 34
    MID_LEVEL_STOP_ACK = 35
    MID_LEVEL_GET_CURRENT_DATA = 36
    MID_LEVEL_GET_CURRENT_DATA_ACK = 37

    DL_INIT = 100
    DL_INIT_ACK = 101
    DL_START = 102
    DL_START_ACK = 103
    DL_STOP = 104
    DL_STOP_ACK = 105
    DL_GET = 109
    DL_GET_ACK = 110
    DL_POWER_MODULE = 111
    DL_POWER_MODULE_ACK = 112
    DL_SYS = 114
    DL_SYS_ACK = 115

    DL_SEND_LIVE_DATA = 106
    DL_SEND_FILE = 107
    DL_SEND_FILE_ACK = 113
    DL_MMI = 108



"""Provides a packet buffer functionality for more async handling of packets and acknowledges"""

from science_mode_4.protocol.packet import Packet
from science_mode_4.protocol.packet_factory import PacketFactory
from science_mode_4.protocol.protocol import Protocol
from science_mode_4.protocol.commands import Commands
from science_mode_4.utils.logger import logger
from .connection import Connection


class PacketBuffer():
    """Class for handling a buffer and provides methods to take care of arriving acknowledges"""

    def __init__(self, conn: Connection, packet_factory: PacketFactory):
        self._buffer: bytes = b""
        # dict with command and packet number as key and open count as value
        self._open_acknowledges: dict[tuple[int, int], int] = {}
        self._connection = conn
        self._packet_factory = packet_factory


    @property
    def connection(self) -> Connection:
        """Getter for connection"""
        return self._connection


    @property
    def buffer(self) -> bytes:
        """Getter for buffer"""
        return self._buffer


    def update_buffer(self):
        """Reads all data from connection and appends to internal buffer"""
        self._buffer += self._connection.read()


    def add_open_acknowledge(self, packet: Packet):
        """Adds packet to list of packets, that are waiting for a acknowledge"""
        key = packet.command + 1, packet.number
        if key in self._open_acknowledges:
            self._open_acknowledges[key] += 1
        else:
            self._open_acknowledges[key] = 1


    def remove_open_acknowledge(self, packet: Packet):
        """Remove packet from list of packets, that are waiting for a acknowledge"""
        key = packet.command + 1, packet.number
        if key in self._open_acknowledges:
            self._open_acknowledges[key] -= 1
        else:
            raise ValueError(f"Remove non existing acknowledge from packet buffer, command {packet.command}, number {packet.number}")


    def print_open_acknowledge(self):
        """Print open acknowledges"""
        logger().info("Open acknowledges")
        for key, value in self._open_acknowledges.items():
            if value != 0:
                logger().info("Command: %s, number: %d, value: %d", Commands(key[0]).name, key[1], value)


    def get_packet_from_buffer(self, do_update_buffer = True) -> Packet | None:
        """Search for a valid packet in buffer and returns found packet. Does adjust internal buffer accordingly.
        Returns None if no valid packet was found
        """
        if do_update_buffer:
            self.update_buffer()

        start_stop = Protocol.find_packet_in_buffer(self._buffer)
        if start_stop is None:
            return None

        packet_data = self._buffer[start_stop[0]: start_stop[1] + 1]
        ack_data = Protocol.extract_packet_data(packet_data)
        # check if we wait for this acknowledge
        key = ack_data[0], ack_data[1]
        wait_ack = self._open_acknowledges.get(key)
        if wait_ack is None:
            if ack_data[0] not in [Commands.DL_SEND_LIVE_DATA]:
                logger().warning("Unexpected acknowledge command: %s, number: %d", Commands(ack_data[0]).name, ack_data[1])
        else:
            self._open_acknowledges[ack_data[0], ack_data[1]] -= 1

        # remove from buffer
        self._buffer = self._buffer[start_stop[0] + start_stop[1] + 1:]
        ack = self._packet_factory.create_packet_with_data(ack_data[0], ack_data[1], ack_data[2])
        return ack


    def clear_buffer(self):
        """Clear internal buffer and buffer from connection"""
        self._connection.clear_buffer()
        self._buffer = b""



"""Provides a class to write values asynchronous to a .csv file"""


import csv
from queue import Empty, Full, Queue
import threading


class CsvHelper:
    """Class for holding values and write to .csv file"""


    def __init__(self, filename: str, header: list[str]):
        self._filename = filename
        self._header = header

        # this queue is used to synchronize data between background and main thread
        self._data_queue = Queue(maxsize=0)
        self._is_running = False


    def start(self):
        """Start background tread"""

        self._is_running = True

        # Create and start the data generator thread (aka background thread)
        data_thread = threading.Thread(target=self._background_task, daemon=True)
        data_thread.start()


    def stop(self):
        """Stop background thread"""
        self._is_running = False


    def append_values(self, package_nr: int, values: list[float], time_delta: int):
        """Appends values, runs in main thread context"""
        try:
            self._data_queue.put_nowait([package_nr] + values + [time_delta])
        except Full:
            # Queue is full, skip this update
            pass


    def _background_task(self):
        with open(self._filename, "w", encoding="utf-8", newline="") as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerow(self._header)

            while self._is_running:
                while self._data_queue.qsize() > 0:
                    try:
                        data = self._data_queue.get_nowait()
                        csv_writer.writerow(data)
                    except Empty:
                        # No new data in the queue
                        pass



"""Provides packet classes for dyscom send live data"""

import struct
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import PacketAck
from .dyscom_types import DyscomElectrodeSample, DyscomSignalType, DyscomPowerLiveDataStatusFlag


class PacketDyscomSendLiveData(PacketAck):
    """Packet for dyscom send live data (this is technically not an acknowledge, but it is handled as such,
    because it is send automatically from device)"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_SEND_LIVE_DATA
        self._number_of_channels = 0
        self._time_offset = 0
        self._samples: list[DyscomElectrodeSample] = []

        if not data is None:
            self._number_of_channels = data[0]
            self._time_offset = int.from_bytes(data[1:5], "big")

            for x in range(self._number_of_channels):
                start_index = 5 + x * 6

                sample = DyscomElectrodeSample()
                sample.value = struct.unpack_from(">f", data[start_index:start_index+4])[0]
                sample.signal_type = DyscomSignalType(data[start_index+4])
                status = data[start_index+5]
                for f in DyscomPowerLiveDataStatusFlag:
                    if status & (1 << f) == 1:
                        sample.status.add(f)

                self._samples.append(sample)


    @property
    def number_of_channels(self) -> int:
        """Getter for number of channels"""
        return self._number_of_channels


    @property
    def time_offset(self) -> int:
        """Getter for time offset"""
        return self._time_offset


    @property
    def samples(self) -> list[DyscomElectrodeSample]:
        """Getter for samples"""
        return self._samples


    @property
    def status_error(self) -> bool:
        """Returns true if in any sample a status flag is set, false otherwise"""
        for x in self._samples:
            if len(x.status) != 0:
                return True
        return False



"""Init file for mid level"""

from .mid_level_current_data import *
from .mid_level_init import *
from .mid_level_layer import *
from .mid_level_stop import *
from .mid_level_types import *
from .mid_level_update import *



"""Test program how to use library without installing the library,
DO NOT USE THIS FILE, USE EXAMPLES INSTEAD"""

from timeit import default_timer as timer

import logging
import sys
import asyncio

from science_mode_4.device_i24 import DeviceI24
from science_mode_4.device_p24 import DeviceP24
from science_mode_4.dyscom.ads129x.ads129x_config_register_1 import Ads129xOutputDataRate, Ads129xPowerMode
from science_mode_4.dyscom.dyscom_get_operation_mode import PacketDyscomGetAckOperationMode
from science_mode_4.dyscom.dyscom_send_live_data import PacketDyscomSendLiveData
from science_mode_4.dyscom.dyscom_types import DyscomFilterType, DyscomGetType, DyscomInitParams, DyscomPowerModulePowerType, DyscomPowerModuleType, DyscomSignalType
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.utils import logger
from science_mode_4.utils.serial_port_connection import SerialPortConnection
from science_mode_4.utils.usb_connection import UsbConnection



async def main() -> int:
    """Main function"""

    logger().disabled = True

    logger().setLevel(logging.DEBUG)
    # devices = SerialPortConnection.list_science_mode_device_ports()
    # connection = SerialPortConnection(devices[0].device)
    devices = UsbConnection.list_science_mode_devices()
    connection = UsbConnection(devices[0])
    # connection = NullConnection()
    connection.open()

    device = DeviceI24(connection)
    await device.initialize()
    # get dyscom layer to call dyscom level commands
    dyscom = device.get_layer_dyscom()

    # call enable measurement power module for measurement
    await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_ON)
    # call init with 4k sample rate and enable signal types
    init_params = DyscomInitParams()
    init_params.filter = DyscomFilterType.PREDEFINED_FILTER_2
    init_params.signal_type = [DyscomSignalType.BI, DyscomSignalType.EMG_1,\
                            DyscomSignalType.EMG_2, DyscomSignalType.BREATHING, DyscomSignalType.TEMPERATURE]
    init_params.register_map_ads129x.config_register_1.output_data_rate = Ads129xOutputDataRate.HR_MODE_4_KSPS__LP_MODE_2_KSPS
    init_params.register_map_ads129x.config_register_1.power_mode = Ads129xPowerMode.HIGH_RESOLUTION
    await dyscom.init(init_params)

    # start dyscom measurement
    await dyscom.start()

    start_time = timer()
    total_count = 0

    # loop for some time
    for x in range(1000):
        # check operation mode from time to time, this function is not waiting for response
        # so we have to handle it by ourself later
        if x % 100 == 0:
            dyscom.send_get_operation_mode()

        live_data_counter = 0
        while True:
            # process all available packages
            ack = dyscom.packet_buffer.get_packet_from_buffer(live_data_counter == 0)
            if ack:
                # because there are multiple get commands, we need to additionally check kind,
                # which is always associated DyscomGetType
                if ack.command == Commands.DL_GET_ACK and ack.kind == DyscomGetType.OPERATION_MODE:
                    om_ack: PacketDyscomGetAckOperationMode = ack
                    print(f"Operation mode {om_ack.operation_mode.name}")
                    # check if measurement is still active
                    if om_ack.result_error != ResultAndError.NO_ERROR:
                        break
                elif ack.command == Commands.DL_SEND_LIVE_DATA:
                    live_data_counter += 1
                    total_count += 1

                    sld: PacketDyscomSendLiveData = ack
                    if sld.status_error:
                        print(f"SendLiveData status error {sld.samples}")
                        break

            else:
                # print(f"Live data acknowledges per iteration {live_data_counter}")
                break

        # await asyncio.sleep(0.01)

    # print stats
    end_time = timer()
    print(f"Samples: {total_count}, duration: {end_time - start_time}, sample rate: {total_count / (end_time - start_time)}")

    # stop measurement
    await dyscom.stop()
    # turn power module off
    await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_OFF)

    connection.close()

    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Init file for utils"""

from .bit_vector import *
from .byte_builder import *
from .connection import *
from .crc16 import *
from .logger import *
from .null_connection import *
from .packet_buffer import *
from .serial_port_connection import *
from .usb_connection import *



"""Provides an example how to use low level layer"""

import asyncio
import sys

from science_mode_4 import DeviceP24
from science_mode_4 import ChannelPoint
from science_mode_4 import SerialPortConnection
from science_mode_4 import LayerLowLevel
from science_mode_4 import Connector, Channel
from science_mode_4 import LowLevelHighVoltageSource, LowLevelMode
from examples.utils.example_utils import ExampleUtils, KeyboardInputThread


def send_channel_config(low_level_layer: LayerLowLevel, connector: Connector):
    """Sends channel update"""
    # device can store up to 10 channel config commands
    for channel in Channel:
        # send_channel_config does not wait for an acknowledge
        low_level_layer.send_channel_config(True, channel, connector,
                                            [ChannelPoint(4000, 20), ChannelPoint(4000, -20),
                                            ChannelPoint(4000, 0)])


async def main() -> int:
    """Main function"""

    # keyboard is our trigger to start specific stimulation
    def input_callback(input_value: str) -> bool:
        """Callback call from keyboard input thread"""
        print(f"Input value {input_value}")

        if input_value == "1":
            send_channel_config(low_level_layer, Connector.GREEN)
        elif input_value == "2":
            send_channel_config(low_level_layer, Connector.YELLOW)
        elif input_value == "q":
            # end keyboard input thread
            return True
        else:
            print("Invalid command")

        return False

    print("Usage: press 1 or 2 to stimulate green or yellow connector or press q to quit")
    # create keyboard input thread for non blocking console input
    keyboard_input_thread = KeyboardInputThread(input_callback)

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get low level layer to call low level commands
    low_level_layer = device.get_layer_low_level()

    # call init low level
    await low_level_layer.init(LowLevelMode.NO_MEASUREMENT, LowLevelHighVoltageSource.STANDARD)

    # now we can start stimulation
    while keyboard_input_thread.is_alive():
        # get new packets from connection
        ack = low_level_layer.packet_buffer.get_packet_from_buffer()
        if ack:
            # do something with packet ack
            # here we print that an acknowledge arrived
            print(ack)

        await asyncio.sleep(0.1)

    # wait until all acknowledges are received
    await asyncio.sleep(0.5)
    # call stop low level
    await low_level_layer.stop()

    # close serial port connection
    connection.close()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides classes for general Reset"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.protocol.types import ResultAndError


class PacketGeneralReset(Packet):
    """Packet for general Reset"""


    def __init__(self):
        super().__init__()
        self._command = Commands.RESET


class PacketGeneralResetAck(PacketAck):
    """Packet for general Reset acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.RESET_ACK
        self._result_error = ResultAndError.NO_ERROR

        if not data is None:
            self._result_error = ResultAndError(data[0])


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error
